#! /usr/bin/env/python 3.1
#
# Script that runs each bat file in each CWE directory.
# The bat files contain instructions for compiling the test cases
# in that directory.
#
# This script must be run in a visual studio command prompt.
#

import sys, glob, os, re

# add parent directory to search path so we can use py_common
sys.path.append("..")

import py_common

goannacc = "goannacc"

import platform

print platform.system()
cwd = os.getcwd()

if platform.system() == 'Windows':
	win_folder = "goanna-central-windows-release-2.4.0-9292"
	goannacc = os.path.join(cwd, win_folder, "bin", "goannacc.exe")
	goannacc += " \"--rcfile=" + os.path.join(cwd, win_folder, "goannaccrc") + "\""

print goannacc


goanna_table = {}
goanna_table['ARR-inv-index-ptr-pos'] = [120, 121, 126, 129, 131, 805]
goanna_table['ARR-inv-index-ptr'] = [120, 121, 126, 129, 131, 805]
goanna_table['ARR-inv-index-pos'] = [120, 121, 126, 129, 131, 805]
goanna_table['ARR-inv-index'] = [120, 121, 126, 129, 131, 805]
goanna_table['ARR-neg-index'] = [124, 127]
goanna_table['ATH-cmp-float'] = [480]
goanna_table['ATH-cmp-unsign-neg'] = []
goanna_table['ATH-cmp-unsign-pos'] = []
goanna_table['ATH-div-0-assign'] = [369]
goanna_table['ATH-div-0-cmp-aft'] = [369]
goanna_table['ATH-div-0-cmp-bef'] = [369]
goanna_table['ATH-div-0-interval'] = [369]
goanna_table['ATH-div-0-pos'] = [369]
goanna_table['ATH-div-0-unchk-param'] = [369]
goanna_table['ATH-div-0'] = [369]
goanna_table['ATH-inc-bool'] = [480]
goanna_table['ATH-neg-check-nonneg'] = [129]
goanna_table['ATH-neg-check-pos'] = []
goanna_table['ATH-overflow-cast'] = [190, 191, 194, 195, 196, 197, 680]
goanna_table['ATH-shift-bounds-32'] = []
goanna_table['ATH-shift-bounds-64'] = []
goanna_table['ATH-shift-bounds-zero'] = []
goanna_table['ATH-shift-neg'] = []
goanna_table['ATH-sizeof-by-sizeof'] = [480]
goanna_table['COP-alloc-ctor'] = [401, 404]
goanna_table['COP-assign-op-ret'] = []
goanna_table['COP-assign-op-self'] = []
goanna_table['COP-assign-op'] = [401, 404]
goanna_table['COP-copy-ctor'] = [401, 404]
goanna_table['COP-dealloc-dtor'] = [401, 404]
goanna_table['COP-dtor-throw'] = []
goanna_table['COP-dtor'] = [401, 404]
goanna_table['COP-init-order'] = [456]
goanna_table['COP-member-uninit'] = [456]
goanna_table['COP-ctor-dtor-leak'] = [401]
goanna_table['CPU-ctor-call-virt'] = []
goanna_table['CPU-delete-throw'] = []
goanna_table['CPU-dtor-call-virt'] = []
goanna_table['CPU-malloc-class'] = []
goanna_table['CPU-nonvirt-dtor'] = []
goanna_table['CST-local'] = []
goanna_table['CST-member-ret'] = []
goanna_table['CST-member'] = []
goanna_table['CST-param'] = []
goanna_table['EXP-cond-assign'] = [481]
goanna_table['EXP-dangling-else'] = [483]
goanna_table['EXP-loop-exit'] = []
goanna_table['EXP-loop-var-mod'] = []
goanna_table['EXP-loop-var-uninit'] = []
goanna_table['EXP-loop-var-unmod'] = []
goanna_table['EXP-main-ret-int'] = []
goanna_table['EXP-null-stmt'] = [483]
goanna_table['FPT-arith-address'] = [480]
goanna_table['FPT-arith'] = [480]
goanna_table['FPT-cmp-null'] = [480]
goanna_table['FPT-literal'] = []
goanna_table['FPT-misuse'] = [480]
goanna_table['ITR-end-cmp-aft'] = []
goanna_table['ITR-end-cmp-bef'] = []
goanna_table['ITR-invalidated'] = [119, 672]
goanna_table['ITR-mismatch-alg'] = []
goanna_table['ITR-postfix'] = []
goanna_table['ITR-store'] = []
goanna_table['ITR-uninit'] = [457]
goanna_table['LIB-buf-arr'] = [119, 120, 121]
goanna_table['LIB-buf-lit'] = [119, 120, 121]
goanna_table['LIB-buf-size'] = [119, 120, 121, 122]
goanna_table['LIB-fn-unsafe'] = [242, 252, 394, 477, 676]
goanna_table['LIB-return-const'] = [252, 394]
goanna_table['LIB-return-error'] = [252, 394]
goanna_table['LIB-return-leak'] = [252, 394]
goanna_table['LIB-return-neg'] = [252, 394]
goanna_table['LIB-return-null'] = [252, 253, 394, 665, 690]
goanna_table['MEM-delete-array-op'] = [404, 762]
goanna_table['MEM-delete-op'] = [404, 762]
goanna_table['MEM-double-free'] = [415]
goanna_table['MEM-free-op'] = [404, 762]
goanna_table['MEM-free-some'] = [401, 476]
goanna_table['MEM-free-variable'] = [590]
goanna_table['MEM-lose-assign'] = [401, 772]
goanna_table['MEM-malloc-arith'] = []
goanna_table['MEM-malloc-diff-type'] = []
goanna_table['MEM-malloc-sizeof-ptr'] = [467]
goanna_table['MEM-malloc-sizeof'] = []
goanna_table['MEM-stack-global-field'] = [466]
goanna_table['MEM-stack-global'] = [466]
goanna_table['MEM-stack-param-ref'] = [466]
goanna_table['MEM-stack-param'] = [466]
goanna_table['MEM-stack-ref'] = [562]
goanna_table['MEM-stack'] = [562]
goanna_table['MEM-use-free-all'] = [416]
goanna_table['MEM-use-free-some'] = [416]
goanna_table['OPT-postfix'] = []
goanna_table['POR-alloc-size'] = []
goanna_table['POR-alloc-sizeof'] = []
goanna_table['POR-imp-cast-const'] = []
goanna_table['POR-imp-cast-ptr-ath'] = []
goanna_table['POR-imp-cast-subscript'] = []
goanna_table['POR-imp-cast-ternary'] = []
goanna_table['PTR-arith-field'] = []
goanna_table['PTR-arith-var'] = [120]
goanna_table['PTR-cmp-str-lit'] = [597]
goanna_table['PTR-null-assign-fun-pos'] = [476]
goanna_table['PTR-null-assign-pos'] = [476]
goanna_table['PTR-null-assign'] = [476]
goanna_table['PTR-null-cmp-aft'] = [476]
goanna_table['PTR-null-cmp-bef-fun'] = [476]
goanna_table['PTR-null-cmp-bef'] = [476]
goanna_table['PTR-null-literal-pos'] = [476, 587]
goanna_table['PTR-null-pos-assign'] = [476]
goanna_table['PTR-unchk-global'] = []
goanna_table['PTR-unchk-param-pos'] = [822]
goanna_table['PTR-unchk-param-some'] = [822]
goanna_table['PTR-unchk-param'] = [822]
goanna_table['PTR-uninit-pos'] = [457, 665, 824]
goanna_table['PTR-uninit'] = [457, 665, 824]
goanna_table['RED-case-reach'] = []
goanna_table['RED-cmp-always'] = [571]
goanna_table['RED-cmp-never'] = [570]
goanna_table['RED-compound-casts'] = [480]
goanna_table['RED-cond-const-assign'] = [481, 570, 571]
goanna_table['RED-cond-const-expr'] = [570, 571]
goanna_table['RED-cond-const'] = [570, 571]
goanna_table['RED-cond-var-always'] = [571]
goanna_table['RED-cond-var-never'] = [570]
goanna_table['RED-dead'] = [561]
goanna_table['RED-expr'] = []
goanna_table['RED-local-hides-global'] = []
goanna_table['RED-local-hides-local'] = []
goanna_table['RED-local-hides-member'] = []
goanna_table['RED-local-hides-param'] = []
goanna_table['RED-no-effect'] = [482]
goanna_table['RED-self-assign'] = [480]
goanna_table['RED-unused-param'] = [563]
goanna_table['RED-unused-val-ptr'] = [401, 563]
goanna_table['RED-unused-val'] = [563]
goanna_table['RED-unused-var-all'] = [563]
goanna_table['SEC-array-access'] = [129]
goanna_table['SEC-command-injection'] = [78]
goanna_table['SEC-external-control'] = [15]
goanna_table['SEC-format-string'] = [134]
goanna_table['SEC-memory-allocation'] = [194, 195, 680, 789]
goanna_table['SEC-memory-copy'] = [194, 195]
goanna_table['SEC-path-traversal'] = [23, 36]
goanna_table['SEC-process-control'] = [114]
goanna_table['SEC-search-path'] = [427]
goanna_table['SEC-string-manipulation'] = []
goanna_table['SEM-const-call'] = []
goanna_table['SEM-const-global'] = []
goanna_table['SEM-pure-call'] = []
goanna_table['SEM-pure-global'] = []
goanna_table['SPC-order'] = [696]
goanna_table['SPC-return'] = [758]
goanna_table['SPC-uninit-arr-all'] = [457]
goanna_table['SPC-uninit-struct-field'] = [457]
goanna_table['SPC-uninit-struct'] = [457]
goanna_table['SPC-uninit-var-all'] = [457]
goanna_table['SPC-uninit-var-some'] = [457]
goanna_table['TST-promo-bit-not'] = []
goanna_table['TST-promo-bit'] = []
goanna_table['TST-promo-overflow'] = []
goanna_table['TST-promo'] = []

check_table = {}
for i in range(1000):
	check_table[i] = []

for key in goanna_table:
	val = goanna_table[key]
	for v in val:
		check_table[v].append(key)

def one_level_deep(file_name):
	return os.path.join(os.path.basename(os.path.dirname(file_name)), os.path.basename(file_name))

db = os.path.join(cwd, "goanna.db")

def run_example_tool(cwe_folder):
	"""
	This method is called from the run_analysis method.  It is called for
	each matching file.  Files are matched against the glob expression
	specified in main.

	When this method is called, the script will have changed to the directory 
	where the batch file exists.
	"""
	# CWE_FOLDER format = CWE<NUM>_desc
	cwe_num = cwe_folder[3:6]
	if cwe_num[-1] == '_':
		cwe_num = cwe_num[:-1]
	cwe_int = int(cwe_num)
	equiv_goanna_checks = check_table[cwe_int]
	if len(equiv_goanna_checks) is 0:
		#skip cwe's that we don't have a check for
		return
	checks = " " + " ".join(['--check=' + check for check in equiv_goanna_checks])
	plain_folder_contents = py_common.find_files_in_dir(cwe_folder, "CWE.*[0-9]\.c.*")
	plain_folder_contents.sort()
	ipa_folder_contents = py_common.find_files_in_dir(cwe_folder, "CWE.*[0-9][a-z]\.c.*")
	ipa_folder_contents.sort()
	# don't need to group the individual test cases.
	test_file_groups = [[one_level_deep(f)] for f in plain_folder_contents]
	ipa_groups = {}
	# group the files like [abc_12b.c, abc_12a.c] because they represent one test case
	for f in ipa_folder_contents:
		key = re.sub(r'[a-z]\.c.*$', r'', f)
		if not ipa_groups.has_key(key):
			ipa_groups[key] = []
		ipa_groups[key].append(one_level_deep(f))
	for key in ipa_groups:
		new_list = ipa_groups[key]
		new_list.sort()
		new_list.reverse()
		test_file_groups.append(new_list)
	for group in test_file_groups:
		pretty_base = re.sub(r'[a-z]?\.c.*', r'', os.path.basename(group[0]))
		prettyname = re.sub(cwe_folder + "__", r'', pretty_base)
		#db = os.path.join(cwe_folder, pretty_base + ".db")
		file_num = len(group)
		sys.stderr.write("Running CWE" + cwe_num + " - test : " + prettyname + " (contains " + str(file_num) + " files)\r\n")
		#run goanna on all the files in the group
		for c_file in group:
			command = goannacc + " --timeout=0 --nc --output-format=\"%ABSFILE%:%LINENO%: warning: Goanna[%CHECKNAME%] %FUNCTION% - %MESSAGE%%EOL%%TRACE%%EOL%\" -I.. -I../testcasesupport/ --parse-error-log=- -include sys/socket.h -include netinet/in.h " + c_file
			if not c_file.endswith(".c"):
				command += " --c++"
			command += checks
			#if file_num > 1: # Some files are interprocedural in one file
			command += " --ipa \"--db=" + db + "\""
			sys.stderr.write("Running " + command + "\r\n")
			try:
				py_common.run_commands([command], use_shell=True)
			except:
				sys.stderr.write("Command failed\r\n")
		# clear call_graph from DB so we don't get conflicts from other test groups
		#   containing functions with the same names
		clear = "echo \"delete from call_graph;\" | sqlite3 " + db
		py_common.run_commands([clear], use_shell=True)

	return

	#command = "goannacc --nc --output-format=\"%ABSFILE%:%LINENO%: warning: Goanna[%CHECKNAME%] %FUNCTION% - %MESSAGE%%EOL%%TRACE%%EOL%\" -I ../../testcasesupport/ --parse-error-log=- -include sys/socket.h -include netinet/in.h " + c_file
	#if not c_file.endswith(".c"):
		#command += " --c++"
	#command += checks

	#sys.stderr.write("Running " + command + "\r\n")
	#py_common.run_commands([command], use_shell=True)

if __name__ == '__main__':

	# First, the run_analysis method finds files according to the 
	# glob regex.  Next, for each file the
	# method finds, it will execute the run_example_tool method.
	py_common.run_analysis("testcases/CWE*", run_example_tool)
